#####################################################################################################
# Python script for extracting elevation values from lidar data for a string of points              #
# Released under the terms of GPLv3 by acrallycentral.com                                           #
# Version 1.3 - 06/12/2024                                                                          #
# To install dependencies: run following command                                                    #
# > conda install conda-forge::pdal gpxpy geopandas shapely numpy pandas pyproj matplotlib          #
#####################################################################################################

import gpxpy
import geopandas as gpd
from shapely.geometry import Point, LineString
import numpy as np
import pandas as pd
import logging, pdal, json, os, sys
import matplotlib.pyplot as plt
import configparser, glob
from shapely.ops import unary_union
from scipy.spatial import cKDTree

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s', datefmt='%H:%M:%S')

def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)

def load_config():
    """Load configuration from INI file"""
    config = configparser.ConfigParser(inline_comment_prefixes='#')
    application_path = os.path.dirname(sys.executable) if getattr(sys, 'frozen', False) else os.path.dirname(os.path.abspath(__file__))
    config_path = resource_path('GPX-to-points-string-with-elevations.ini')
    logging.info(f"Looking for config file at: {config_path}")
    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file not found: {config_path}")
    config.read(config_path)
    return config, application_path

def get_file_paths(config, application_path):
    """Get file paths from config"""
    input_file = os.path.abspath(os.path.join(application_path, config.get('Files', 'INPUT_FILE')))
    input_base = os.path.splitext(os.path.basename(input_file))[0]
    return {
        'input_file': input_file,
        'copc_file': os.path.abspath(os.path.join(application_path, config.get('Files', 'COPC_FILE'))),
        'copc_dir': os.path.abspath(os.path.join(application_path, config.get('Files', 'COPC_DIR'))),
        'output_gpkg': os.path.join(application_path, config.get('Output', 'OUTPUT_GPKG').format(input_base=input_base)),
        'output_csv': os.path.join(application_path, config.get('Output', 'OUTPUT_CSV').format(input_base=input_base)),
        'output_png': os.path.join(application_path, config.get('Output', 'OUTPUT_PNG').format(input_base=input_base))
    }

def check_files_and_folders(file_paths, multi_file):
    """Check for the presence of required files and folders"""
    if not os.path.exists(file_paths['input_file']):
        raise FileNotFoundError(f"Input GPX file not found: {file_paths['input_file']}")

    if multi_file:
        if not os.path.isdir(file_paths['copc_dir']):
            raise FileNotFoundError(f"COPC directory not found: {file_paths['copc_dir']}")
        copc_files = glob.glob(os.path.join(file_paths['copc_dir'], '*.copc.laz'))
        if not copc_files:
            raise FileNotFoundError(f"No COPC files found in directory: {file_paths['copc_dir']}")
    else:
        if not os.path.exists(file_paths['copc_file']):
            raise FileNotFoundError(f"COPC file not found: {file_paths['copc_file']}")

    logging.info("All required files and folders are present.")

def create_points_at_equal_distances(line, distance):
    """Create points at equal distances along a line"""
    points = []
    current_distance = 0
    line_length = line.length
    while current_distance <= line_length:
        points.append(line.interpolate(current_distance))
        current_distance += distance
    if points[-1].distance(Point(line.coords[-1])) > 1e-8:
        points.append(Point(line.coords[-1]))
    return points

def idw_interpolation_vectorized(tree, points, values, target_points, num_points, power=2):
    """Perform vectorized IDW interpolation for multiple target points"""
    distances, indices = tree.query(target_points, k=num_points)
    # Avoid division by zero
    distances = np.maximum(distances, 1e-10)
    weights = 1 / (distances ** power)
    weighted_values = values[indices] * weights
    return np.sum(weighted_values, axis=1) / np.sum(weights, axis=1)

def get_elevations_from_memory_vectorized(target_points, lidar_points, num_points, use_idw):
    """Get elevations for multiple points from in-memory LiDAR data using vectorized operations"""
    points = np.column_stack((lidar_points['X'], lidar_points['Y']))
    elevations = lidar_points['Z']
    
    # Create KD-tree for efficient nearest neighbor search
    tree = cKDTree(points)
    
    if use_idw:
        return idw_interpolation_vectorized(tree, points, elevations, target_points, num_points)
    else:
        # For simple averaging
        distances, indices = tree.query(target_points, k=num_points)
        return np.mean(elevations[indices], axis=1)

def create_height_profile(distances, elevations, output_png):
    """Create and save height profile plot"""
    plt.figure(figsize=(12, 6))
    plt.plot(distances, elevations)
    plt.title('Height Profile')
    plt.xlabel('Distance (m)')
    plt.ylabel('Elevation (m)')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig(output_png, format='png', dpi=300)
    plt.close()
    logging.info(f"Height profile saved as {output_png}")

def get_copc_files(copc_file, copc_dir, multi_file):
    """Get list of COPC files based on configuration"""
    return glob.glob(os.path.join(copc_dir, '*.copc.laz')) if multi_file else [copc_file]

def create_buffer_polygon(gdf_copc_crs, buffer_distance):
    """Create a buffer polygon around the GPX track"""
    buffered_geometries = gdf_copc_crs.geometry.buffer(buffer_distance)
    return unary_union(buffered_geometries)

def read_lidar_points(copc_files, buffer_polygon, classified):
    """Read LiDAR points within the buffer polygon into memory"""
    all_points = []
    
    for copc_file in copc_files:
        bounds = buffer_polygon.bounds
        json_pipeline = [
            {
                "type": "readers.copc",
                "filename": copc_file,
                "polygon": buffer_polygon.wkt
            }
        ]
        
        if classified:
            json_pipeline.append({
                "type": "filters.range",
                "limits": "Classification[2:2],Classification[11:11],Classification[17:17]"
            })
        
        pipeline = pdal.Pipeline(json.dumps(json_pipeline))
        pipeline.execute()
        
        arrays = pipeline.arrays
        if len(arrays) > 0:
            points = arrays[0]
            all_points.append(points)
    
    if not all_points:
        raise ValueError("No LiDAR points found within the buffer polygon")
    
    return np.concatenate(all_points)

def process_gpx_with_elevation(input_file, output_gpkg, output_csv, output_png, sample_distance, copc_files, default_crs, include_distance, num_points, use_idw, classified):
    """Main processing function"""
    with open(input_file, 'r') as gpx_file:
        gpx = gpxpy.parse(gpx_file)
    
    track_points = [Point(point.longitude, point.latitude) for track in gpx.tracks for segment in track.segments for point in segment.points]
    logging.info(f"Number of original track points: {len(track_points)}")
    
    if len(track_points) < 2:
        raise ValueError("The GPX file must contain at least two points")
    
    gdf = gpd.GeoDataFrame(geometry=track_points, crs="EPSG:4326")
    
    gdf_copc_crs = gdf.to_crs(default_crs)
    line_copc_crs = LineString(gdf_copc_crs.geometry.tolist())
    logging.info(f"Original track length: {line_copc_crs.length:.2f} meters")
    
    equal_distance_points_copc_crs = create_points_at_equal_distances(line_copc_crs, sample_distance)
    logging.info(f"Number of equally spaced points: {len(equal_distance_points_copc_crs)}")
    
    gdf_points_copc_crs = gpd.GeoDataFrame(geometry=equal_distance_points_copc_crs, crs=default_crs)
    gdf_points = gdf_points_copc_crs.to_crs("EPSG:4326")
    
    buffer_polygon = create_buffer_polygon(gdf_points_copc_crs, 2)  # 2 meter buffer

    logging.info(f"Loading LiDAR data...")
    lidar_points = read_lidar_points(copc_files, buffer_polygon, classified)
    logging.info(f"Loaded {len(lidar_points)} LiDAR points into memory")

    # Convert GeoDataFrame geometries to numpy array of coordinates
    target_points = np.array([[point.x, point.y] for point in gdf_points_copc_crs.geometry])
    
    logging.info("Fetching elevations...")
    elevations = get_elevations_from_memory_vectorized(
        target_points, 
        lidar_points, 
        num_points, 
        use_idw
    )

    gdf_points['elevation'] = elevations
    gdf_points['cumulative_distance'] = gdf_points_copc_crs.distance(gdf_points_copc_crs.shift()).cumsum().fillna(0)
    
    gdf_points.to_file(output_gpkg, driver="GPKG", layer='track_points')
    logging.info(f"Processed GPX file with elevations saved as {output_gpkg}")
    
    csv_data = pd.DataFrame({
        'longitude': gdf_points.geometry.x,
        'latitude': gdf_points.geometry.y,
        'elevation': gdf_points.elevation.round(2),
    })

    if include_distance:
        csv_data['cumulative_distance'] = gdf_points.cumulative_distance.round(2)

    csv_data.to_csv(output_csv, index=False)
    logging.info(f"Processed GPX file with elevations{'and cumulative distances' if include_distance else ''} saved as {output_csv}")

    create_height_profile(gdf_points['cumulative_distance'], csv_data['elevation'], output_png)

if __name__ == "__main__":
    try:
        config, application_path = load_config()
        file_paths = get_file_paths(config, application_path)
        multi_file = config.getboolean('Processing', 'MULTI_FILE')

        check_files_and_folders(file_paths, multi_file)

        copc_files = get_copc_files(file_paths['copc_file'], file_paths['copc_dir'], multi_file)

        process_gpx_with_elevation(
            file_paths['input_file'], file_paths['output_gpkg'], file_paths['output_csv'], file_paths['output_png'],
            config.getfloat('Processing', 'POINTS_DISTANCE'),
            copc_files,
            config.getint('Processing', 'DEFAULT_CRS'),
            config.getboolean('Processing', 'INCLUDE_DISTANCE'),
            config.getint('Processing', 'NUM_POINTS'),
            config.getboolean('Processing', 'USE_IDW'),
            config.getboolean('Processing', 'CLASSIFIED')
        )
    except FileNotFoundError as e:
        logging.error(f"File not found error: {str(e)}")
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}", exc_info=True)
    finally:
        if config.getboolean('Processing', 'WAIT_FOR_KEY', fallback=True):
            print("\nPress any key to exit...")
            input()
        else:
            print("\nProcessing complete. Exiting...")
